unit GeneratorPhp;

interface

uses
  Classes, SysUtils, strutils, Source, Generator;

type

  { TGeneratorPhp }

  TGeneratorPhp = class(TGenerator)
  private
    procedure GenerateProgram(Block: TBlock;  Prog:TProgram);
    procedure GenerateFunction(ParentBlock: TBlock; FunctionDef: TFunction);
    procedure GenerateBlock(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockConst(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBlockFunctions(ParentBlock: TBlock; Block: TBlock);
    procedure GenerateBeginEnd(Block: TBlock; BeginEnd: TBeginEnd);
    procedure GenerateCommand(Block: TBlock; Command: TCommand);
    procedure GenerateIfThenElse(Block: TBlock; IfThenElse: TIfThenElse);
    procedure GenerateWhileDo(Block: TBlock; WhileDo: TWhileDo);
    procedure GenerateRepeatUntil(Block: TBlock; RepeatUntil: TRepeatUntil);
    procedure GenerateForToDo(Block: TBlock; ForToDo: TForToDo);
    procedure GenerateFunctionCall(Block: TBlock; FunctionCall: TFunctionCall);
    procedure GenerateAssignment(Block: TBlock; Assignment: TAssignment);
    procedure GenerateExpression(Block: TBlock; Expression: TExpression);
    procedure GenerateExpressionOperation(Block: TBlock; Expression: TExpressionOperation);
    procedure GenerateExpressionOperand(Block: TBlock; Expression: TExpressionOperand);
    procedure GenerateExpressionBrackets(Block: TBlock; Expression: TExpressionBrackets);
    procedure GenerateBreak(Block: TBlock; BreakCmd: TBreak);
    procedure GenerateReturn(Block: TBlock; Return: TReturn);
    procedure GenerateContinue(Block: TBlock; ContinueCmd: TContinue);
    procedure GenerateValue(Value: TValue);
  public
    procedure Generate; override;
    constructor Create; override;
  end;


implementation

const
  ExpressionOperatorTextPhp: array[TExpressionOperator] of string = ('', '+',
    '-', '*', '/', '/', '%', 'and', 'xor', 'or', '<<',
    '>>', '==', '!=', '<', '>', '<=','>=', '!');

{ TGeneratorPhp }

procedure TGeneratorPhp.GenerateCommand(Block: TBlock; Command: TCommand);
begin
  if Command is TBeginEnd then GenerateBeginEnd(Block, TBeginEnd(Command))
  else if Command is TFunctionCall then GenerateFunctionCall(Block, TFunctionCall(Command))
  else if Command is TAssignment then GenerateAssignment(Block, TAssignment(Command))
  else if Command is TIfThenElse then GenerateIfThenElse(Block, TIfThenElse(Command))
  else if Command is TWhileDo then GenerateWhileDo(Block, TWhileDo(Command))
  else if Command is TRepeatUntil then GenerateRepeatUntil(Block, TRepeatUntil(Command))
  else if Command is TForToDo then GenerateForToDo(Block, TForToDo(Command))
  else if Command is TBreak then GenerateBreak(Block, TBreak(Command))
  else if Command is TContinue then GenerateContinue(Block, TContinue(Command))
  else if Command is TReturn then GenerateReturn(Block, TReturn(Command))
  else if Command is TEmptyCommand then
  else raise Exception.Create('Unsupported command type');
end;

procedure TGeneratorPhp.GenerateIfThenElse(Block: TBlock; IfThenElse: TIfThenElse);
begin
  AddText('if (');
  GenerateExpression(Block, IfThenElse.Expression);
  AddText(' ) ');
  GenerateCommand(Block, IfThenElse.CommandThen);
  if Assigned(IfThenElse.CommandElse) and not (IfThenElse.CommandElse is TEmptyCommand) then begin
    if Copy(Output, Length(Output), 1) <> '}' then AddText(';');
    AddText(' else ');
    GenerateCommand(Block, IfThenElse.CommandElse);
  end;
end;

procedure TGeneratorPhp.GenerateWhileDo(Block: TBlock; WhileDo: TWhileDo);
begin
  AddText('while (');
  GenerateExpression(Block, WhileDo.Expression);
  AddText(' ) ');
  GenerateCommand(Block, WhileDo.Command);
end;

procedure TGeneratorPhp.GenerateRepeatUntil(Block: TBlock;
  RepeatUntil: TRepeatUntil);
var
  I: Integer;
begin
  AddTextLine('while (true)');
  AddTextLine('{');
  Indent := Indent + 1;
  for I := 0 to RepeatUntil.Commands.Count - 1 do begin
    GenerateCommand(Block, TCommand(RepeatUntil.Commands[I]));
    AddTextLine(';');
  end;
  AddText('if (');
  GenerateExpression(Block, RepeatUntil.Expression);
  AddTextLine(') break;');
  Indent := Indent - 1;
  AddText('}');
end;

procedure TGeneratorPhp.GenerateForToDo(Block: TBlock; ForToDo: TForToDo);
begin
  AddText('for (');
  AddText('$' + ForToDo.VariableRef.Name);
  AddText(' = ');
  GenerateExpression(Block, ForToDo.ExpressionFrom);
  AddText(' ; ');
  AddText('$' + ForToDo.VariableRef.Name + ' <= ');
  GenerateExpression(Block, ForToDo.ExpressionTo);
  AddText(' ; ');
  AddText('$' + ForToDo.VariableRef.Name + '++');
  AddText(' ) ');
  GenerateCommand(Block, ForToDo.Command);
end;

procedure TGeneratorPhp.GenerateFunctionCall(Block: TBlock;
  FunctionCall: TFunctionCall);
var
  I: Integer;
begin
  AddText(FunctionCall.FunctionDef.Name);
  if FunctionCall.Params.Count > 0 then begin
    AddText('(');
    for I := 0 to FunctionCall.Params.Count - 1 do
      GenerateExpression(Block, TExpression(FunctionCall.Params[I]));
    AddText(')');
  end;
end;

procedure TGeneratorPhp.GenerateAssignment(Block: TBlock; Assignment: TAssignment);
begin
  AddText('$' + Assignment.Variable.Name);
  AddText(' = ');
  GenerateExpression(Block, Assignment.Expression);
end;

procedure TGeneratorPhp.GenerateExpression(Block: TBlock; Expression: TExpression);
begin
  if Expression is TExpressionOperation then
    GenerateExpressionOperation(Block, TExpressionOperation(Expression))
  else
  if Expression is TExpressionOperand then
    GenerateExpressionOperand(Block, TExpressionOperand(Expression))
  else
  if Expression is TExpressionBrackets then
    GenerateExpressionBrackets(Block, TExpressionBrackets(Expression))
  else raise Exception.Create('Unknown expression class.');
end;

procedure TGeneratorPhp.GenerateExpressionOperation(Block: TBlock;
  Expression: TExpressionOperation);
var
  I: Integer;
begin
  for I := 0 to Expression.Items.Count - 1 do begin
    if I > 0 then begin
      if Expression.Operation = eoAdd then begin
        if Expression.TypeRef.Name = 'string' then AddText('.')
        else AddText(' + ');
      end
      else AddText(' ' + ExpressionOperatorTextPhp[Expression.Operation] + ' ');
    end;
    GenerateExpression(Block, TExpression(Expression.Items[I]));
  end;
end;

procedure TGeneratorPhp.GenerateExpressionOperand(Block: TBlock;
  Expression: TExpressionOperand);
begin
  case Expression.OperandType of
    otFunctionCall: GenerateFunctionCall(Block, Expression.FunctionCall);
    otConstantDirect: GenerateValue(Expression.ConstantDirect.Value);
    otConstantRef: AddText('$' + Expression.ConstantRef.Name);
    otVariableRef: AddText('$' + Expression.VariableRef.Name);
    else raise Exception.Create('Unsupported exception operand type.');
  end;
end;

procedure TGeneratorPhp.GenerateExpressionBrackets(Block: TBlock;
  Expression: TExpressionBrackets);
begin
  AddText('(');
  GenerateExpression(Block, Expression.Expression);
  AddText(')');
end;

procedure TGeneratorPhp.GenerateBreak(Block: TBlock; BreakCmd: TBreak);
begin
  AddText('break');
end;

procedure TGeneratorPhp.GenerateReturn(Block: TBlock; Return: TReturn);
begin
  AddText('return ');
  GenerateExpression(Block, Return.Expression);
end;

procedure TGeneratorPhp.GenerateContinue(Block: TBlock; ContinueCmd: TContinue);
begin
  AddText('continue');
end;

procedure TGeneratorPhp.GenerateValue(Value: TValue);
begin
  if Value is TValueBoolean then begin
    if TValueBoolean(Value).Value then AddText('true') else AddText('false');
  end else if Value is TValueString then AddText('''' + StringReplace(TValueString(Value).Value, '''', '\''', [rfReplaceAll]) + '''')
  else if Value is TValueInteger then AddText(IntToStr(TValueInteger(Value).Value))
  else raise Exception.Create('Unsupported value type.');
end;

procedure TGeneratorPhp.GenerateProgram(Block: TBlock; Prog: TProgram);
begin
  AddTextLine('<?php');
  if Prog.Name <> '' then AddTextLine('// Program ' + Prog.Name);

  GenerateBlock(nil, Prog.SystemBlock);
  GenerateBlock(Block, Prog.Block);
end;

procedure TGeneratorPhp.GenerateFunction(ParentBlock: TBlock;
  FunctionDef: TFunction);
var
  I: Integer;
begin
  AddText('function ' + FunctionDef.Name + '(');
  for I := 0 to FunctionDef.Params.Count - 1 do begin
    AddText('$' + TFunctionParameter(FunctionDef.Params[I]).Name);
    if I > 0 then AddText(', ');
  end;
  AddTextLine(')');
  if FunctionDef.InternalName <> '' then begin
    AddTextLine('{');
    Indent := Indent + 1;
    if FunctionDef.InternalName = 'WriteLn' then AddTextLine('echo($Text."\n");')
    else if FunctionDef.InternalName = 'Write' then AddTextLine('echo($Text);')
    else if FunctionDef.InternalName = 'IntToStr' then AddTextLine('return $Value;')
    else if FunctionDef.InternalName = 'StrToInt' then AddTextLine('return $Value;');
    Indent := Indent - 1;
    AddTextLine('}');
  end else begin
    GenerateBlock(ParentBlock, FunctionDef.Block);
    AddTextLine;
  end;
end;

procedure TGeneratorPhp.GenerateBlock(ParentBlock: TBlock; Block: TBlock);
begin
  GenerateBlockConst(ParentBlock, Block);
  GenerateBlockFunctions(ParentBlock, Block);
  if Block.BeginEnd.Commands.Count > 0 then begin
    GenerateBeginEnd(ParentBlock, Block.BeginEnd);
    AddTextLine;
  end;
end;

procedure TGeneratorPhp.GenerateBlockConst(ParentBlock: TBlock; Block: TBlock);
var
  I: Integer;
  Constant: TConstant;
begin
  for I := 0 to Block.Constants.Count - 1 do begin
    Constant := TConstant(Block.Constants[I]);
    AddText('$' + Constant.Name + ' = ');
    GenerateValue(Constant.Value);
    AddTextLine(';');
  end;
end;

procedure TGeneratorPhp.GenerateBlockFunctions(ParentBlock: TBlock;
  Block: TBlock);
var
  I: Integer;
begin
  for I := 0 to Block.Functions.Count - 1 do begin
    GenerateFunction(ParentBlock, TFunction(Block.Functions[I]));
    AddTextLine;
  end;
end;

procedure TGeneratorPhp.GenerateBeginEnd(Block: TBlock; BeginEnd: TBeginEnd);
var
  I: Integer;
begin
  AddTextLine('{');
  Indent := Indent + 1;
  for I := 0 to BeginEnd.Commands.Count - 1 do begin
    GenerateCommand(Block, TCommand(BeginEnd.Commands[I]));
    AddTextLine(';');
  end;
  Indent := Indent - 1;
  AddText('}');
end;

procedure TGeneratorPhp.Generate;
begin
  Output := '';
  GenerateProgram(Prog.SystemBlock, Prog);
end;

constructor TGeneratorPhp.Create;
begin
  inherited;
  Name := 'PHP';
  FileExt := '.php';
end;

end.

